<isol>

/**
 * @file main.cpp
 */

/*
 * Copyright (C) Gemfony scientific UG (haftungsbeschraenkt)
 *
 * Contact: contact [at] gemfony (dot) com
 *
 * This file is part of the vectorAdd program.
 *
 * vectorAdd is free software: you can redistribute and/or modify it under
 * the terms of version 3 of the GNU Affero General Public License
 * as published by the Free Software Foundation.
 *
 * vectorAdd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with the Geneva library. If not, see <http://www.gnu.org/licenses/>.
 *
 * For further information on Gemfony scientific, visit http://www.gemfony.com
 */

#include <iostream>
#include <algorithm>

// This will force OpenCL C++ classes to raise exceptions
// rather than to use an error code
#define __CL_ENABLE_EXCEPTIONS

// OpenCL includes
#include "LOCAL_CL/cl.hpp"

/********************************************************************************/
/**
 * The OpenCL source
 */
std::string OCLSource =
"__kernel void vector_add ("
"  const __global float *a"
"  , const __global float *b"
"  , __global float * result"
"){"
"  size_t pos = get_global_id(0);"
"  result[pos] = a[pos] + b[pos];"
"}";

/********************************************************************************/
/**
 * Returns the platforms available on this computer
 */
std::vector<cl::Platform> getPlatforms() {
   std::vector<cl::Platform> platforms;

   // Extract the platforms
   cl::Platform::get(&platforms);

   // Let the audience know
   return platforms;
}

/********************************************************************************/
/**
 * Emits the devices of a specified device type for a given platform
 */
std::vector<cl::Device> getDevices(const cl::Platform& platform, const cl_device_type& t) {
   std::vector<cl::Device> devices;
   
   // Extract a vector of devices
   platform.getDevices(t, &devices);

   // Let the audience know about the platforms
   for(std::size_t nd=0; nd<devices.size(); nd++) {
     std::cout << "Found device with name " << devices[nd].getInfo<CL_DEVICE_NAME>() << std::endl;
   }
   
   // Let the audience know
   return devices;
}

/********************************************************************************/
/**
 * The main function
 */
const std::size_t NENTRIES = 1000;
int main(int argc, char** argv)
{
  // Some OpenCL declarations
  std::vector<cl::Platform> platforms;
  std::vector<cl::Device> devices;
  cl::Context context;
  cl::CommandQueue queue;
  cl::Program::Sources sources;
  cl::Program program;
  cl::Kernel kernel;
  cl::Buffer a_buffer, b_buffer, result_buffer;

  // Initialize and fill some input and output buffers
  cl_float *a      = new cl_float[NENTRIES];
  cl_float *b      = new cl_float[NENTRIES];
  cl_float *result = new cl_float[NENTRIES];

  for(std::size_t i=0; i<NENTRIES; i++) {
    a[i]      = (cl_float)i;
    b[i]      = (cl_float)i;
    result[i] = 0.f; // result[i] should be 2*i
  }

  try {
    // Retrieve devices
    platforms = getPlatforms();
    std::cout << "Found " << platforms.size() << " platforms" << std::endl;
    
    // Use the first platform only
    devices = getDevices(platforms[0], CL_DEVICE_TYPE_GPU);
    
    // Build a context for the devices
    context = cl::Context(devices);
    
    // Build a queue  for the first device -- we will ignore the others
    cl::CommandQueue queue(context, devices[0]);
    
    // Create a program object
    sources = cl::Program::Sources(1, std::make_pair(OCLSource.c_str(), OCLSource.length()));
    program = cl::Program(context, sources);

    // Build the program for our devices. Note that we can pass compiler options here
    // We could just as well pass a DEFINE into the kernel here
    program.build(devices,"-cl-fast-relaxed-math");

    // Create a kernel object
    kernel = cl::Kernel(program, "vector_add");

    // Create buffers for the kernel arguments
    a_buffer = cl::Buffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, NENTRIES*sizeof(cl_float), (void *)a);
    b_buffer = cl::Buffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, NENTRIES*sizeof(cl_float), (void *)b);
    result_buffer = cl::Buffer(context, CL_MEM_WRITE_ONLY, NENTRIES*sizeof(cl_float)); // No data is copied to the device

    // Set the kernel arguments
    kernel.setArg(0, a_buffer);
    kernel.setArg(1, b_buffer);
    kernel.setArg(2, result_buffer);

    // Submit the kernel and wait for its termination
    cl::Event event;
    queue.enqueueNDRangeKernel(kernel, cl::NullRange, cl::NDRange(NENTRIES), cl::NullRange, NULL, &event); 
    event.wait();

    // Extract the result data
    queue.enqueueReadBuffer(result_buffer, CL_TRUE, 0, NENTRIES*sizeof(cl_float), (void *)result);
    
    // Output a sub-set of the retrieved data
    for(std::size_t i=0; i < std::min<std::size_t>(10,NENTRIES); i++) {
      std::cout << result[i] << " // Should be " << 2*i << std::endl;
    }
  } catch(cl::Error& err) {
    std::cout 
      << err.what() << std::endl
      << err.err() << std::endl;

    // If we ran into an error, it was likely due to a build failure
    std::cout << "Build log:\t"
	      << program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(devices[0]) 
	      << std::endl;
  }

  // Get rid of the buffers
  delete [] a;
  delete [] b;
  delete [] result;

  return 0;
}
